package clients;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Paths;
import java.util.ArrayList;

import org.jgrapht.alg.shortestpath.BellmanFordShortestPath;
import operations.Audit;
import operations.Transport;
import operations.Deliver;
import operations.Path;
import operations.Attend;
import roads.State;
import roads.Location;
import roads.ModelSynthesiser;
import roads.OWCModel;
import roads.Planner;
import roads.Verification;
import roads.Roads;
import roads.Street;
import roads.Task;
import verification.RoadModel;
import verification.RoadProperty;
import verification.RoadResult;

public class RoadsTest {

	public static String projectPath = Paths.get("").toAbsolutePath().toString()+"/scenarios/";

	public static String emergencyScenarioPath = projectPath+"emergencyScenario/";
	public static String deliverScenarioPath = projectPath+"deliverScenario/";
	public static String parametricDeliverScenarioPath = projectPath+"parametricDeliverScenario/";	
	public static String prismBinLocation = "/Users/kjohnson/prism-4.5-osx64/bin";



	private static void writefile(String pathAndName,String contents)
	{

		File file = new File(pathAndName);
		try {

			file.getParentFile().mkdirs();
			file.createNewFile();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}


		try (PrintWriter out = new PrintWriter(pathAndName)) 
		{
			out.println(contents);
		} catch (FileNotFoundException e) 
		{
			e.printStackTrace();
		}	
	}


	private static void script(ArrayList<String> params,String pathAndName,ArrayList<String> files,String propertyFile,String modelCheckerCommand,String modelCheckerLocation)
	{


		String content = "# run by typing . ./runscenario.sh";

		//-param p03=0.0:1.0,
		String pCommands = "";
		if(!params.isEmpty())
		{
			pCommands = "-param ";
			for(String p : params)
			{
				pCommands += p+"=0.0:1.0,";
			}
			pCommands = pCommands.substring(0, pCommands.length()-1);
		}


		for(String name : files)
		{
			content += "\n "+modelCheckerLocation+"/"+modelCheckerCommand+" "+name + " "+propertyFile+" "+pCommands+" >> scenariooutput.txt\n";
			content += "\n\n >> scenariooutput.txt";

		}
		content += "\ngrep -e 'Parsing model file' -e 'Time' -e 'Result' scenariooutput.txt";
		writefile(pathAndName,content);
	}

	public static void deliverShortestPathParametricScenario(Roads roads,Deliver deliveraudit)
	{


		System.out.println("Deliver Parametric (Shortest Path) Scenario");
		long start = System.currentTimeMillis();
		String out = "";
		Planner planner = roads.planner;
		ModelSynthesiser synth = roads.synth;
		Verification prob = roads.pmc;

		ArrayList<Path> routes = planner.shortestPathResolve(deliveraudit);
		String propertyName = parametricDeliverScenarioPath+"/properties.pctl";

		ArrayList<String> files = new ArrayList<String>();


		writefile(propertyName,synth.taskSuccessFormula(deliveraudit));


		roads.sigma.parameters.add("pSoftware");
		roads.sigma.parameters.add("pTraining");
		roads.sigma.parameters.add("pRetry");		

		for(int i=0;i<routes.size();i++)
		{
			Path route = (Path) routes.get(i);		
			RoadModel model = synth.synthesise(deliveraudit, route);
			RoadProperty succ = new RoadProperty("success",synth.taskSuccessFormula(deliveraudit));
			RoadProperty cost = new RoadProperty("success",synth.taskExpectedCostFormula(deliveraudit));

			String fileName = parametricDeliverScenarioPath+"/parametricDeliver"+i+".pm";
			files.add(fileName);
			model.body = "//"+route.path.toString()+"\n"+model.body;
			if(roads.sigma.parameters.isEmpty())//parameters in model
			{
				RoadResult pmcResult = prob.verify(model, succ,cost);
				out+=pmcResult+"\n";
			}
			writefile(fileName,model.body);

		}
		System.out.println(out);

		script(synth.sigma.parameters,parametricDeliverScenarioPath+"/runDeliverScenario.sh",files,propertyName,"prism",prismBinLocation);
		//long stop = System.currentTimeMillis();
		//long duration = stop - start;

		//return duration;
	}


	public static long emergencyAllPathsScenario(Roads roads,Task emergency)
	{

		long start = System.currentTimeMillis();

		Planner planner = roads.planner;
		ModelSynthesiser synth = roads.synth;
		Verification prob = roads.pmc;

		System.out.println("Emergency All Paths Scenario");
		ArrayList<Path> emergencyRoutes = planner.allDirectedPathsResolve(emergency);
		String propertyName = emergencyScenarioPath+"/properties.pctl";

		writefile(propertyName,synth.taskSuccessFormula(emergency)+"\n"+synth.taskExpectedCostFormula(emergency));

		ArrayList<String> files = new ArrayList<String>();

		for(int i=0;i<emergencyRoutes.size();i++)
		{			
			Path path = (Path) emergencyRoutes.get(i);			
			RoadModel model = synth.synthesise(emergency, path);
			String fileName = emergencyScenarioPath+"/emergencyroute"+i+".pm";
			files.add(fileName);
			model.body = "//"+path.path.toString()+"\n"+model.body;

			RoadProperty succ = new RoadProperty("success",synth.taskSuccessFormula(emergency));
			RoadProperty cost = new RoadProperty("cost",synth.taskExpectedCostFormula(emergency));

			RoadResult pmcResult = prob.verify(model, succ,cost);
			long nStates = prob.prism.getBuiltModel().getNumStates();
			long nTrans = prob.prism.getBuiltModel().getNumTransitions();

			writefile(fileName,model.body);

			output(path, (double)pmcResult.results.get(succ).getResult(),(double)pmcResult.results.get(cost).getResult(),nStates,nTrans);
		}
		script(synth.sigma.parameters,emergencyScenarioPath+"/runEmergencyScenario.sh",files,propertyName,"prism",prismBinLocation);
		long stop = System.currentTimeMillis();
		long duration = stop - start;
		return duration;
	}

	private static void output(Path pi,double succ,double cost,long nStates,long nTrans)
	{

		boolean paper_output=false;
		String s = String.format("%.5g%n", succ);
		s = s.replaceAll("\n", "");		
		String c = String.format("%.5g%n", cost);
		c = c.replaceAll("\n", "");
		String out = "";
		for(Location l : pi.path.getVertexList())
		{

			out+=paper_output?l.label+"\\to ":l.label+" -> ";
		}
		out = out.substring(0,out.length()-4);
		if(paper_output)
		{
			System.out.println("$"+out+"$"+" & "+s+" & "+c+" & "+nStates+" & "+nTrans+"\\\\");
		}
		else
		{
			System.out.println(out+"  "+s+"  "+c+"  "+nStates+"  "+nTrans);
		}

	}

	public static long deliverAllPathsScenario(Roads roads,Deliver deliver)
	{


		long start = System.currentTimeMillis();
		System.out.println("Deliver (All Paths) Scenario");

		Planner planner = roads.planner;
		ModelSynthesiser synth = roads.synth;
		Verification prob = roads.pmc;

		ArrayList<Path> paths = planner.allDirectedPathsResolve(deliver);
		String propertyName = deliverScenarioPath+"/properties.pctl";

		ArrayList<String> files = new ArrayList<String>();

		writefile(propertyName,synth.taskSuccessFormula(deliver)+"\n"+synth.taskExpectedCostFormula(deliver));


		for(int i=0;i<paths.size();i++)
		{			
			Path path = (Path) paths.get(i);		
			RoadModel model = synth.synthesise(deliver, path);
			RoadProperty succ = new RoadProperty("success",synth.taskSuccessFormula(deliver));
			RoadProperty cost = new RoadProperty("cost",synth.taskExpectedCostFormula(deliver));

			String fileName = deliverScenarioPath+"/deliverroute"+i+".pm";
			files.add(fileName);
			model.body = "//"+path.path.toString()+"\n"+model.body;
			RoadResult pmcResult = prob.verify(model, succ,cost);
			long nStates = prob.prism.getBuiltModel().getNumStates();
			long nTrans = prob.prism.getBuiltModel().getNumTransitions();

			output(path, (double)pmcResult.results.get(succ).getResult(),(double)pmcResult.results.get(cost).getResult(),nStates,nTrans);		
			writefile(fileName,model.body);

		}
		long stop = System.currentTimeMillis();
		long duration = stop - start;
	//	System.out.println("Duration: "+duration+"ms");
		script(synth.sigma.parameters,deliverScenarioPath+"/rundeliverscenario.sh",files,propertyName,"prism",prismBinLocation);
		return duration;
	}



	//--------------------------------------------------------------------------------------------------------------
	private static void deliverAuditShortestPathScale()
	{
		ArrayList<State> ams = new ArrayList<State>();

		int startK = 10;
		int endK = 100;
		for(int k = startK;k<=endK;k+=10)
		{
			System.out.println("G(n,M) = G("+k+","+k*4+")");
			ams.add(RandomGraphs.randomAbstractModel(k, k*4));
		}

		for(State am : ams)
		{

			//System.out.println("Graph: ")
			Roads roads = new Roads(am);
			Planner planner = roads.planner;
			ModelSynthesiser synth = roads.synth;
			Verification pmc = roads.pmc;

			//location of supply and where we must deliver the resources.
			Location current = am.supply;

			boolean done = false;
			Location demand = am.locations[am.n-1];
			int x = 1;
			while(!done)
			{				
				BellmanFordShortestPath bsp = (BellmanFordShortestPath) planner.sp;	
				if((current==demand)||(bsp.findPathBetween(am.graph,current, demand)==null))
				{
					x++; 
					demand = am.locations[am.n-x];

				}
				else 
				{
					done=true;
				}
			}


			//generate the delivery + audit task.
			OWCModel owcmodel = new OWCModel();
			Transport deliver = new Transport(owcmodel,current,demand,10.0);
			Audit audit = new Audit(owcmodel,demand, 0.99, 0.98);
			Deliver deliverAndAudit = new Deliver(owcmodel,deliver,audit);

			//statistics about the graph we generated at random
			ArrayList<Path> routes = planner.shortestPathResolve(deliverAndAudit);


			System.out.println("Deliver + Audit Task");
			System.out.println("Graph has V = "+am.graph.vertexSet().size()+" E = "+am.graph.edgeSet().size());
			System.out.println("Number of routes: "+routes.size());


			int count=0;
			long numberOfStates = 0;
			long numberOfTransitions = 0;
			//double modelCheckingTimeForSucc=0.0;
			long start = System.currentTimeMillis();
			for(Path route : routes)
			{
				count++;
				RoadModel model = synth.synthesise(deliverAndAudit, route);
				RoadProperty success = new RoadProperty("Success",synth.taskSuccessFormula(deliverAndAudit));
				RoadProperty cost = new RoadProperty("Expected Cost",synth.taskExpectedCostFormula(deliverAndAudit));
				RoadResult r = pmc.verify(model,success,cost);
				numberOfStates += r.numberOfStates;
				numberOfTransitions += r.numberOfTransitions;

				//modelCheckingTimeForSucc+=
				//prism.Result fromPrismResult = (prism.Result) r.results.get(success).getResult();


				//Prism. (r.results.get(success).getResult())


				//	System.out.println(r);
				//String succstr = (String.format("%.5g%n", r.results.get(success).getResult()));
				//String coststr = (String.format("%.5g%n", r.results.get(cost).getResult()));
				//output(route);
				//System.out.println("$s_"+i+"$& $"+output((Route)s)+"$  & "+succstr+" & "+coststr+"\\\\");

			}
			long end = System.currentTimeMillis();

			long duration = end-start;
			System.out.println("allPathsResolve took: "+duration+"ms");

			double avgStates = numberOfStates/count;
			double avgTrans = numberOfTransitions/count;
			System.out.println("average number of states of model: "+avgStates+" average number of transitions of model: "+avgTrans);
		}
	}
	//--------------------------------------------------------------------------------------------------------------
	private static void deliverAuditAllPathsScale()
	{
		ArrayList<State> ams = new ArrayList<State>();

		int startK = 10;
		int endK = 100;
		for(int k = startK;k<=endK;k+=10)
		{
			int eK= k*4;
			System.out.println("G(n,M) = G("+k+","+eK+")");
			ams.add(RandomGraphs.randomAbstractModel(k, eK));
		}

		for(State am : ams)
		{
			for(Street s : am.graph.edgeSet())
			{
				if (am.graph.getEdgeTarget(s) == am.graph.getEdgeSource(s))
				{
					System.out.println("has self loops");
				}
			}

			//System.out.println("Graph: ")
			Roads roads = new Roads(am);
			Planner planner = roads.planner;
			ModelSynthesiser synth = roads.synth;
			Verification pmc = roads.pmc;

			//location of supply and where we must deliver the resources.
			Location current = am.supply;

			boolean done = false;
			Location demand = am.locations[am.n-1];
			int x = 1;
			while(!done)
			{				
				BellmanFordShortestPath bsp = (BellmanFordShortestPath) planner.sp;	
				if((current==demand)||(bsp.findPathBetween(am.graph,current, demand)==null))
				{
					x++;
					demand = am.locations[am.n-x];
				}
				else 
				{
					done=true;
				}
			}

			//generate the delivery + audit task.
			OWCModel owcModel = new OWCModel();
			Transport deliver = new Transport(owcModel,current,demand,10.0);
			Audit audit = new Audit(owcModel,demand, 0.99, 0.98);
			Deliver deliverAndAudit = new Deliver(owcModel,deliver,audit);

			//statistics about the graph we generated at random
			ArrayList<Path> routes = planner.allDirectedPathsResolve(deliverAndAudit);


			System.out.println("Deliver + Audit Task");
			System.out.println("Graph has V = "+am.graph.vertexSet().size()+" E = "+am.graph.edgeSet().size());
			System.out.println("Number of routes: "+routes.size());


			int count=0;
			long numberOfStates = 0;
			long numberOfTransitions = 0;
			//double modelCheckingTimeForSucc=0.0;
			long start = System.currentTimeMillis();
			for(Path route : routes)
			{
				count++;
				RoadModel model = synth.synthesise(deliverAndAudit, route);
				RoadProperty success = new RoadProperty("Success",synth.taskSuccessFormula(deliverAndAudit));
				RoadProperty cost = new RoadProperty("Expected Cost",synth.taskExpectedCostFormula(deliverAndAudit));
				RoadResult r = pmc.verify(model,success,cost);
				numberOfStates += r.numberOfStates;
				numberOfTransitions += r.numberOfTransitions;

				//modelCheckingTimeForSucc+=
				//prism.Result fromPrismResult = (prism.Result) r.results.get(success).getResult();


				//Prism. (r.results.get(success).getResult())


				//	System.out.println(r);
				//String succstr = (String.format("%.5g%n", r.results.get(success).getResult()));
				//String coststr = (String.format("%.5g%n", r.results.get(cost).getResult()));
				//output(route);
				//System.out.println("$s_"+i+"$& $"+output((Route)s)+"$  & "+succstr+" & "+coststr+"\\\\");

			}
			long end = System.currentTimeMillis();

			long duration = end-start;
			System.out.println("allPathsResolve took: "+duration+"ms");

			double avgStates = numberOfStates/count;
			double avgTrans = numberOfTransitions/count;
			System.out.println("average number of states of model: "+avgStates+" average number of transitions of model: "+avgTrans);
		}

	}

	public static void main(String[] args) throws Exception 
	{	

		OWCModel owcModel = new OWCModel();

		State city = Metropolis.cityModel();

		System.out.println("Number of input graph vertices: "+city.graph.vertexSet().size());
		System.out.println("Number of input graph edges: "+city.graph.edgeSet().size());

		Roads roads = new Roads(city);

		Task attend = new Attend(owcModel,Metropolis.emergency,Metropolis.d,0.98);
		Transport transport = new Transport(owcModel,Metropolis.a,Metropolis.f,0.95);
		Audit audit = new Audit(owcModel,Metropolis.f, 0.98, 0.96);

		Deliver deliver = new Deliver(owcModel,transport,audit);

		System.out.println(deliverAllPathsScenario(roads,deliver)+"ms\n");
		System.out.println(emergencyAllPathsScenario(roads,attend)+"ms\n");
		deliverShortestPathParametricScenario(roads,deliver);


	}

}
