package roads;


import java.util.List;

import operations.Audit;
import operations.Transport;
import operations.Deliver;
import operations.Path;
import operations.Attend;
import verification.RoadModel;

public class ModelSynthesiser
{

	static public String state = "state";

	public State sigma;	
	//---------------------------------------------------------------
	private String parameterAssignment(String p,double value)
	{
		if (sigma.parameters.contains(p))
		{
			return "";
		}

		return "= "+value;
	}
	//---------------------------------------------------------------
	private String parameterAssignment(Street s)
	{
		String p = sigma.probEdgeConst(s.getSource(), s.getTarget());
		if (sigma.parameters.contains(p))
		{
			return "";
		}
		return "= "+s.probability;
	}

	//---------------------------------------------------------------
	public String taskSuccessFormula(Task task)
	{
		return "P=?[F(tasksuccess=1)]";
	}
	//---------------------------------------------------------------
	//independent of the task being completed....
	public String taskExpectedCostFormula(Task task)
	{
		//return "//map traversal expected cost\nR=?[F((l=SUCC)|(l=FAIL))]";
		return "R=?[C]";
	}
	//---------------------------------------------------------------

	//---------------------------------------------------------------
	private String linesc(String text)
	{

		return this.linecr(text+";");
	}
	//---------------------------------------------------------------
	private String linecr(String text)
	{
		return text+"\n";	
	}
	//---------------------------------------------------------------
	public ModelSynthesiser(State am)
	{
		this.sigma = am;
	}	
	//---------------------------------------------------------------
	private boolean strategyIncludesEdge(Path r,Street s)
	{
		Path path = (Path) r;
		List<Street> edges = path.path.getEdgeList();
		return edges.contains(s);		 
	}
	//---------------------------------------------------------------
	private String costEdgeConst(Street s)
	{
		String out = s.getSource().label+s.getTarget().label;
		return "cost"+out;
	}
	//---------------------------------------------------------------
	private String cityModule(String predForBeginningTraversal,Task task,Path r)
	{
		String body  = "";
		String probConstants = "";

		String costConstants = "";
		String edgeConstants = "";
		String stateConstants = "";
		String rewardRules = this.linecr("rewards");

		String strategyCmds = "";

		for(Street street : this.sigma.graph.edgeSet())
		{

			Location i = street.getSource();
			Location j = street.getTarget();
			String iLabel = i.label;
			String jLabel = j.label;
			double cost = street.getWeight();


			boolean value  = this.strategyIncludesEdge(r,street);

			//No traversal back to previous state if at the last state in the route.
			boolean value2=false;
			if (r.path.getEndVertex() != street.getTarget())
				value2 = this.strategyIncludesEdge(r,this.sigma.graph.getEdge(street.getTarget(), street.getSource()));


			edgeConstants +=this.linesc("const bool "+sigma.edgeConst(i, j)+" = "+value);//+" | "+value2);
			probConstants += this.linesc("const double "+sigma.probEdgeConst(i,j)+this.parameterAssignment(street));
			costConstants += this.linesc("const double "+this.costEdgeConst(street)+" = "+cost);
			rewardRules += this.linesc("["+iLabel+jLabel+"] true : "+this.costEdgeConst(street));

		}
		rewardRules += this.linecr("endrewards");
		for(int i=0;i<sigma.n;i++)
		{
			stateConstants+=linesc("const int "+sigma.locations[i].label+" = "+(i+1));
		}

		String startLocation = "";
		if ((task instanceof Deliver))
		{
			startLocation = linesc("const int startLocation = supply");
		}
		else
			if (task instanceof Attend)
			{
				startLocation = linesc("const int startLocation = emergency");
			}

		body += this.linecr(stateConstants);
		body += this.linecr(probConstants);

		body += this.linecr(costConstants);
		body += this.linecr(edgeConstants);

		body+=linesc("const int supply = "+sigma.supply.label);
		body+=linesc("const int emergency = "+sigma.dispatch.label);


		if((task instanceof Deliver))
		{
			Deliver dTask = (Deliver) task;
			body+=linesc("const int SUCC = "+dTask.transport.destination.label);
		}
		else
			if((task instanceof Transport))
			{
				Transport dTask = (Transport) task;
				body+=linesc("const int SUCC = "+dTask.destination.label);
			}
			else
				if(task instanceof Attend)
				{
					Attend eTask = (Attend) task;
					body+=linesc("const int SUCC = "+eTask.destination.label);

				}

		body+=linesc("const int FAIL = "+(sigma.n+1));

		body += rewardRules;

		body+=linecr(startLocation);


		body+=linecr("module "+sigma.name+"map");


		body+=linesc(state+" : [0.."+(sigma.n+1)+"] init 0");

		String guard = "";
		body+= linesc("[start]"+predForBeginningTraversal+"&(state=0) -> (state'=startLocation)");
		for(Street e : sigma.graph.edgeSet())
		{
			Location s = sigma.graph.getEdgeSource(e);
			Location t = sigma.graph.getEdgeTarget(e);



			for(Street es : sigma.graph.edgesOf(s))
			{				
				Location tes = sigma.graph.getEdgeTarget(es);

				if(tes!=s)
				{
					String pred = (tes == t)?sigma.edgeConst(s, tes):"";//"!"+am.edgeConst(s, tes);
					guard+= pred;
				}
			}

			String nextState = sigma.probEdgeConst(s, t)+":("+state+"'="+t.label+") + (1 - "+sigma.probEdgeConst(s, t)+"):("+state+"'=FAIL)";
			strategyCmds+=this.linesc("["+s.label+t.label+"] ("+state+"="+s.label+")&"+guard+" -> "+nextState);

			guard = "";
		}

		body+=strategyCmds;
		body+=linecr("endmodule");
		return body;
	}

	//---------------------------------------------------------------
	private RoadModel synthesise(Task task)
	{
		return this.synthesise(task, null);		
	}
	//---------------------------------------------------------------
	public RoadModel synthesise(Task task, Path r) 
	{
		String body = this.linecr(this.linecr("dtmc"));
		body+=this.taskModule(task,r);
		RoadModel model = new RoadModel(body);
		return model;
	}
	//---------------------------------------------------------------

	//---------------------------------------------------------------
	private String taskModule(Task task,Path r)
	{

		if(task instanceof Deliver)
		{
			String citydtmc = this.cityModule("(deliverfo=1)&(deliverfw=1)",task, r);
			String taskbody = this.linecr(citydtmc);
			taskbody += deliverAuditModule((Deliver)task);			
			return taskbody;
		}		
		if(task instanceof Attend)
		{
			String citydtmc = this.cityModule("(emergencyfo=1)&(emergencyfw=1)",task, r);
			String taskbody = this.linecr(citydtmc);
			taskbody += emergencyModule((Attend)task);			
			return taskbody;
		}
		if (task instanceof Audit)
		{
			return auditModule((Audit)task);
		}

		return "";
	}
	//---------------------------------------------------------------
	private String deliverAuditModule(Deliver daTask) 
	{		


		boolean withRetry = true;
		Transport dtask = daTask.transport;
		Audit task = daTask.audit;


		String tasksucc = "";
		tasksucc+=this.linecr("module taskSuccess");

		tasksucc+=this.linesc("tasksuccess : [0..1] init 0");
		tasksucc+=this.linesc("[] (tasksuccess=0)&(deliverfo=1)&(deliverfw=1)&(deliverfc=1)&(auditfo=1)&(auditfo=1)&(auditfc=1)&(state!=FAIL) -> (tasksuccess'=1)");
		tasksucc+=this.linesc("[] (tasksuccess=1) -> true");
		tasksucc+=this.linecr("endmodule");



		//deliver opportunity
		String deliverOpp = "";
		deliverOpp+=this.linecr("module deliveropportunity");
		deliverOpp+=this.linesc("deliverfo : [0..1] init 0");
		deliverOpp+=this.linesc("[] (deliverfo=0) -> (deliverfo'=1)");
		deliverOpp+=this.linecr("endmodule");

		//deliver willingness
		String deliverWill = "";
		deliverWill += this.linesc("const double deliverwp"+this.parameterAssignment("deliverwp", dtask.model.willingness));
		deliverWill += this.linecr("module deliverwillingness");
		deliverWill += this.linesc("deliverfw : [0..2] init 0");
		deliverWill += this.linesc("[] (deliverfw=0)&(deliverfo=1) -> deliverwp:(deliverfw'=1) + (1-deliverwp):(deliverfw'=2)");
		deliverWill += this.linecr("endmodule");

		String deliverCap = "";
		deliverCap+=this.linecr("module delivercapabilities");
		deliverCap+=this.linesc("deliverfc : [0..1] init 0");
		deliverCap+=this.linesc("[] (deliverfc=0)&(deliverfo=1)&(deliverfw=1)&(state=SUCC) -> (deliverfc'=1)");
		deliverCap+=this.linecr("endmodule");


		String deliverBody =  this.linecr(deliverOpp)+this.linecr(deliverWill)+this.linecr(deliverCap);



		String auditOpp = "";
		auditOpp += this.linecr("module auditopportunity");
		auditOpp += this.linesc("auditfo : [0..1] init 0");
		auditOpp += this.linesc("[] (auditfo=0)&(deliverfo=1)&(deliverfw=1)&(deliverfc=1) -> (auditfo'=1)");
		auditOpp += this.linecr("endmodule");

		String auditWill = this.linesc("const double auditwp"+this.parameterAssignment("auditwp", task.model.willingness));
		auditWill += this.linecr("module auditwillingness");
		auditWill += this.linesc("auditfw : [0..2] init 0");
		auditWill += this.linesc("[] (auditfw=0)&(auditfo=1) -> auditwp:(auditfw'=1) + (1-auditwp):(auditfw'=2)");
		auditWill += this.linecr("endmodule");



		String auditSwCap = "";
		if(withRetry)
		{
			auditSwCap+=this.linesc("const double pSoftware"+this.parameterAssignment("pSoftware", task.pSoftware));
			auditSwCap+=this.linesc("const double pRetry"+this.parameterAssignment("pRetry", task.pRetry));
			auditSwCap+=this.linecr("module capabilitySoftware");
			auditSwCap+=this.linesc("fcsw : [0..3] init 0");
			auditSwCap+=this.linesc("[sw] (fcsw=0)&(auditfw=1) -> pSoftware:(fcsw'=1) + (1-pSoftware):(fcsw'=2)");
			auditSwCap+=this.linesc("[retry] (fcsw=2)&(auditfw=1) -> pRetry:(fcsw'=0) + (1 - pRetry):(fcsw'=3)");
			auditSwCap+=this.linecr("endmodule");
		}
		else
		{
			auditSwCap+=this.linesc("const double pSoftware"+this.parameterAssignment("pSoftware", task.pSoftware));		
			auditSwCap+=this.linecr("module capabilitySoftware");
			auditSwCap+=this.linesc("fcsw : [0..2] init 0");
			auditSwCap+=this.linesc("[sw] (fcsw=0)&(auditfw=1) -> pSoftware:(fcsw'=1) + (1-pSoftware):(fcsw'=2)");		
			auditSwCap+=this.linecr("endmodule");
		}


		String auditTrainCap = this.linesc("const double pTraining"+this.parameterAssignment("pTraining", task.pTraining));		
		auditTrainCap+=this.linecr("module capabilityTraining");
		auditTrainCap+=this.linesc("fcTraining : [0..2] init 0");
		auditTrainCap+=this.linesc("[train] (fcTraining=0)&(fcsw=1)&(auditfw=1)  -> pTraining:(fcTraining'=1) + (1-pTraining):(fcTraining'=2)");
		auditTrainCap+=this.linecr("endmodule");


		String auditCap = "";		
		auditCap+= this.linecr("module auditcapabilities");
		auditCap+=this.linesc("auditfc : [0..1] init 0");
		auditCap+= this.linesc("[] (auditfc=0)&(fcsw=1)&(fcTraining=1) -> (auditfc'=1)");
		auditCap+=this.linecr("endmodule");


		String auditBody =  this.linecr(auditOpp)+this.linecr(auditWill)+
				this.linecr(auditCap)+this.linecr(auditSwCap)+this.linecr(auditTrainCap);


		return tasksucc+deliverBody+auditBody;
	}	//---------------------------------------------------------------
	//---------------------------------------------------------------	
	private String auditModule(Audit task) 
	{		
		String body = "";
		return body;
	}
	//---------------------------------------------------------------
	private String emergencyModule(Attend task) {

		String body = "";

		String tasksucc = "";
		tasksucc+=this.linecr("module taskSuccess");

		tasksucc+=this.linesc("tasksuccess : [0..1] init 0");
		tasksucc+=this.linesc("[] (tasksuccess=0)&(emergencyfo=1)&(emergencyfw=1)&(emergencyfc=1)&(state!=FAIL) -> (tasksuccess'=1)");
		tasksucc+=this.linesc("[] (tasksuccess=1) -> true");
		tasksucc+=this.linecr("endmodule");


		String emergencyOpp = "";
		emergencyOpp += this.linecr("module emergencyopportunity");
		emergencyOpp += this.linesc("emergencyfo : [0..1] init 0");
		emergencyOpp += this.linesc("[] (emergencyfo=0) -> (emergencyfo'=1)");
		emergencyOpp += this.linecr("endmodule");

		String emergencyWill = this.linesc("const double emergencywp"
				+this.parameterAssignment("emergencywp", task.model.willingness));
		emergencyWill += this.linecr("module emergencywillingness");
		emergencyWill += this.linesc("emergencyfw : [0..2] init 0");
		emergencyWill += this.linesc("[] (emergencyfw=0)&(emergencyfo=1) -> "
				+ "emergencywp:(emergencyfw'=1) + (1-emergencywp):(emergencyfw'=2)");
		emergencyWill += this.linecr("endmodule");

		String equipCap = this.linesc("const double pEquip"+this.parameterAssignment("pEquip", task.pEquip));
		equipCap+=this.linecr("module pEquipment");
		equipCap+=this.linesc("equipfc : [0..2] init 0");
		equipCap+=this.linesc("[equip] (equipfc=0)&(emergencyfo=1)&(emergencyfw=1)&(state=SUCC) -> pEquip:(equipfc'=1) + (1-pEquip):(equipfc'=2)");
		//equipCap+=this.linesc("[] (fcEquip=1) -> true");
		//equipCap+=this.linesc("[] (fcEquip=2) -> true");
		equipCap+=this.linecr("endmodule");

		//equipCap+=this.linesc("label \"hascap\" = (fc=1)");
		String emergencyCap = this.linecr("module emergencycapabilities");
		emergencyCap+=this.linesc("emergencyfc : [0..1] init 0");
		emergencyCap+= this.linesc("[] (emergencyfc=0)&(equipfc=1) -> (emergencyfc'=1)");
		emergencyCap+=this.linecr("endmodule");

		body += tasksucc+emergencyOpp+emergencyWill+equipCap+emergencyCap;

		return body;
	}


	//---------------------------------------------------------------


}

















