package roads;

import java.util.ArrayList;
import java.util.List;
import org.jgrapht.GraphPath;
import org.jgrapht.alg.interfaces.ShortestPathAlgorithm;
import org.jgrapht.alg.shortestpath.AllDirectedPaths;
import org.jgrapht.alg.shortestpath.BellmanFordShortestPath;
import operations.Transport;
import operations.Deliver;
import operations.Path;
import operations.Attend;


public class Planner{

	static boolean LOG=true;
	private State sigma;
	public ShortestPathAlgorithm sp;
	private AllDirectedPaths adp;

	//---------------------------------------------------------------------------------------------------------
	public Planner(State sigma)
	{
		this.sigma = sigma;
		this.sp  = new BellmanFordShortestPath(sigma.graph);
		this.adp  = new AllDirectedPaths(sigma.graph);
	}
	//---------------------------------------------------------------------------------------------------------
	public ArrayList<Path> shortestPathResolve(Task task)
	{
		ArrayList<Path> paths = new ArrayList<Path>();

		if (task instanceof Deliver)
		{
			Deliver da = (Deliver) task;
			GraphPath<Location,Street> p = sp.getPath(sigma.supply, da.transport.destination);
			Path aPath = new Path(p);
			paths.add(aPath);
		}
		if (task instanceof Transport)
		{
			Transport deliver = (Transport) task;
			GraphPath<Location,Street> p = sp.getPath(sigma.supply, deliver.destination);
			Path aPath = new Path(p);
			paths.add(aPath);
		}
		else
			if(task instanceof Attend)
			{
				Attend emergency = (Attend) task;				
				GraphPath<Location,Street> p = sp.getPath(sigma.dispatch, emergency.destination);
				Path aPath = new Path(p);
				paths.add(aPath);
			} 
		return paths;
	}
	//---------------------------------------------------------------------------------------------------------
	public ArrayList<Path> allDirectedPathsResolve(Task task) 
	{
		List<GraphPath<Location,Street>>  list  = null;
		if (task instanceof Deliver)
		{
			Deliver da = (Deliver) task;
			list = adp.getAllPaths(sigma.supply, da.transport.destination, true, null);

		}
		else
			if(task instanceof Attend)
			{
				Attend emergency = (Attend) task;
				list = adp.getAllPaths(sigma.dispatch, emergency.destination, true, null);			
			} 

		ArrayList<Path> paths = new ArrayList<Path>();
		for(GraphPath<Location,Street> p : list)
		{
			paths.add(new Path(p));
		}
		//computes a route through the disaster...
		return paths;
	}
}